
#-------------------------------------------------------------------------------
# OFORTH : ARM 32bits assembler definitions
#-------------------------------------------------------------------------------

#define LIT         #
#define ALIGNED     .align 16;
#define CELLSIZE    4
#define CELLSHIFT   #2
#define BITSIZE     #32

#define ip          pc
#define sp          sp
#define rs          sp
#define ds          r6
#define thread      r7

#define rf          r0
#define rfl         r0
#define rfw         r0
#define TOS         r5
#define r1          r1
#define r2          r2
#define r2l         r2
#define r3          r3
#define rrem        r3
#define r4          r5

#define at(x)       [x]
#define atm(x,y)    [x, y]
#define atn(x,n)    [x, #(CELLSIZE * n)]
#define atb(x,n)    [x, #n]
#define at1(x)      [x, #CELLSIZE]
#define at2(x)      [x, #(CELLSIZE * 2)]
#define at3(x)      [x, #(CELLSIZE * 3)]
#define at4(x)      [x, #(CELLSIZE * 4)]
#define at5(x)      [x, #(CELLSIZE * 5)]
#define at6(x)      [x, #(CELLSIZE * 6)]
#define at7(x)      [x, #(CELLSIZE * 7)]
#define at8(x)      [x, #(CELLSIZE * 8)]
#define at9(x)      [x, #(CELLSIZE * 9)]
#define at10(x)     [x, #(CELLSIZE * 10)]
#define at11(x)     [x, #(CELLSIZE * 11)]
#define at12(x)     [x, #(CELLSIZE * 12)]
#define at13(x)     [x, #(CELLSIZE * 13)]
#define at14(x)     [x, #(CELLSIZE * 14)]
#define at15(x)     [x, #(CELLSIZE * 15)]
#define at16(x)     [x, #(CELLSIZE * 16)]
#define at17(x)     [x, #(CELLSIZE * 17)]
#define at18(x)     [x, #(CELLSIZE * 18)]
#define at19(x)     [x, #(CELLSIZE * 19)]
#define at20(x)     [x, #(CELLSIZE * 20)]
#define at21(x)     [x, #(CELLSIZE * 21)]

#define bt1(x)      [x, #-CELLSIZE]
#define bt2(x)      [x, #-(CELLSIZE * 2)]
#define bt3(x)      [x, #-(CELLSIZE * 3)]
#define bt4(x)      [x, #-(CELLSIZE * 4)]
#define bt5(x)      [x, #-(CELLSIZE * 5)]
#define bt6(x)      [x, #-(CELLSIZE * 6)]
#define bt7(x)      [x, #-(CELLSIZE * 7)]

#define jmp         b
#define je          beq
#define jne         bne
#define jae         bpl
#define jge         bge
#define jg          bgt
#define jl          blt
#define jle         ble
#define jbe         bls
#define call        bl
#define ret         pop { lr }; bx lr; 

#define push(x)     push {x}
#define pushimm(x)  ldr r4, x; push {r4}
#define pushm(x)    ldr r4, x; push {r4}
#define pop(x)      pop {x}
#define ldm(r, m)   ldr r, m
#define ldg(r, g)   ldr r, =g
#define stm(r, m)   str r, m
#define mov(x, y)   mov y, x
#define movimm(x,y) mov r4, x; str r4, y
#define mvb(x,y)    movb x, y
#define add(x,y)    adds y, x
#define addm(m, x)  ldr r4, m; adds x, r4
#define sub(x,y)    subs y, x
#define subm(i,m)   ldr r4, m; subs r4, i; str r4, m
#define mul(x)      imul x
#define inc(x)      add x, x, #1
#define dec(x)      sub x, x, #1
#define cmp(x,y)    cmp y, x
#define cmpm(i,m)   ldr r4, m; cmp r4, i
#define cmpb(x,y)   cmpb x, y
#define and(x,y)    and  y, y, x
#define or(x,y)     orr y, x
#define orm(x,y)    ldr r4, y; orr r4, r4, x; str r4, y
#define orb(x,y)    orb  x, y
#define xor(x,y)    xorl  x, y
#define test(x,y)   tst x, y
#define testm(m,i)  ldr r4, m; tst r4, i
#define testb(x,y)  testb y, x
#define shr(x, y)   lsr x, y
#define shl(x, y)   lsl x, y
#define sar(x, y)   asr y, x
#define sal(x, y)   lsl y, x
#define lea(x, y)   leal x, y

#define call1(f, x)       mov(x, r0); bl SYM(f)
#define call2(f, x, y)    mov(x, r0); mov(y, r1); bl SYM(f)
#define call3(f, x, y, z) mov(x, r0); mov(y, r1); mov(z, r2); bl SYM(f)

#define SAVEREG     push %ebp; push %ebx; push %edi; push %esi
#define CPARAM1     at5(sp)
#define CPARAM2     at6(sp)
#define RESTREG     pop %esi; pop %edi; pop %ebx; pop %ebp

#define CPAR1       at1(sp)
#define CPAR2       at2(sp)
#define CPAR3       at3(sp)

#define indirectJmp(x, l1, l2)     jmp *x
#define indirectCall(x, l, l1, l2) mov lr, pc; bx r3;

// #define indirectJmp(x,l1,l2)    call l2; l1: pause; jmp l1; l2: movl x, (%esp); ret
// #define indirectCall(x,l,l1,l2) call l; l: indirectJmp(x, l1, l2)

#ifdef __FLOAT_ENABLE__

#define checkUndef(x, l)  mov(at(x), r3); check(r3); jne l ; \
                          cmp($0xFFF80000, at1(x)); je LFloatError

#define tofloat(x)    leal at2(x), x
#define addrfloat(x, y)   leal at2(x), y
#define fadd(x, y, r) fldl (x); faddl (y); fstpl (r)
#define fsub(x, y, r) fldl (x); fsubl (y); fstpl (r)
#define fmul(x, y, r) fldl (x); fmull (y); fstpl (r)
#define fdiv(x, y, r) fldl (x); fdivl (y); fstpl (r)
#define fsqrt(x, r)   fldl (x); fsqrt; fstpl (r)
#define fln(x, r)     fldln2; fldl (x); fyl2x; fstpl (r)

#define fpow(x, y, r) fldl (y); fldl (x); fyl2x; fld1; fld %st(1); fprem; \
                      f2xm1; faddp; fscale; fxch %st(1); fstp %st(0); fstpl (r)

#define fcos(x, r)    fldl (x); fcos; fstpl (r)
#define fsin(x, r)    fldl (x); fsin; fstpl (r)
#define ftan(x, r)    fldl (x); fptan; fincstp; fstpl (r)
#define fatan(x, r)   fldl (x); fld1; fpatan; fstpl (r)

#define facos(x, r)   fldl (x); fld %st(0); fld1; fsubp; fld1; fadd %st(2); \
                      fmulp; fsqrt; fabs; fxch %st(1); fpatan; fstpl (r)

#define fasin(x, r)   fldl (x); fld %st(0); fld1; fsubp; fld1; fadd %st(2); \
                      fmulp; fsqrt; fabs; fpatan; fstpl (r)

#define frand         mov($SYM(N_Float_rand), r3); call callNative

#endif

